@@ -34,8 +34,8 @@
 #include <glib/gi18n.h>
 #include <glib-object.h>
 
+#include <xcb/xcb.h>
 #include <X11/Xlib.h>
-#include <X11/Xatom.h>
 
 #include "gdm-common.h"
 #include "gdm-display.h"
@@ -48,7 +48,6 @@
 
 #include "gdm-launch-environment.h"
 #include "gdm-dbus-util.h"
-#include "gdm-xerrors.h"
 
 #define INITIAL_SETUP_USERNAME "gnome-initial-setup"
 #define GNOME_SESSION_SESSIONS_PATH DATADIR "/gnome-session/sessions"
@@ -76,7 +75,8 @@ struct GdmDisplayPrivate
 
         guint                 finish_idle_id;
 
-        Display              *x11_display;
+        xcb_connection_t     *xcb_connection;
+        int                   xcb_screen_number;
 
         GDBusConnection      *connection;
         GdmDisplayAccessFile *user_access_file;
@@ -292,32 +292,17 @@ gdm_display_create_authority (GdmDisplay *self)
 }
 
 static void
-setup_xhost_auth (XHostAddress              *host_entries,
-                  XServerInterpretedAddress *si_entries)
+setup_xhost_auth (XHostAddress              *host_entries)
 {
-        si_entries[0].type        = "localuser";
-        si_entries[0].typelength  = strlen ("localuser");
-        si_entries[1].type        = "localuser";
-        si_entries[1].typelength  = strlen ("localuser");
-        si_entries[2].type        = "localuser";
-        si_entries[2].typelength  = strlen ("localuser");
-
-        si_entries[0].value       = "root";
-        si_entries[0].valuelength = strlen ("root");
-        si_entries[1].value       = GDM_USERNAME;
-        si_entries[1].valuelength = strlen (GDM_USERNAME);
-        si_entries[2].value       = "gnome-initial-setup";
-        si_entries[2].valuelength = strlen ("gnome-initial-setup");
-
         host_entries[0].family    = FamilyServerInterpreted;
-        host_entries[0].address   = (char *) &si_entries[0];
-        host_entries[0].length    = sizeof (XServerInterpretedAddress);
+        host_entries[0].address   = "localuser\0root";
+        host_entries[0].length    = sizeof ("localuser\0root");
         host_entries[1].family    = FamilyServerInterpreted;
-        host_entries[1].address   = (char *) &si_entries[1];
-        host_entries[1].length    = sizeof (XServerInterpretedAddress);
+        host_entries[1].address   = "localuser\0" GDM_USERNAME;
+        host_entries[1].length    = sizeof ("localuser\0" GDM_USERNAME);
         host_entries[2].family    = FamilyServerInterpreted;
-        host_entries[2].address   = (char *) &si_entries[2];
-        host_entries[2].length    = sizeof (XServerInterpretedAddress);
+        host_entries[2].address   = "localuser\0gnome-initial-setup";
+        host_entries[2].length    = sizeof ("localuser\0gnome-initial-setup");
 }
 
 gboolean
@@ -331,8 +316,8 @@ gdm_display_add_user_authorization (GdmDisplay *self,
         gboolean              res;
 
         int                       i;
-        XServerInterpretedAddress si_entries[3];
         XHostAddress              host_entries[3];
+        xcb_void_cookie_t         cookies[3];
 
         g_return_val_if_fail (GDM_IS_DISPLAY (self), FALSE);
 
@@ -382,14 +367,25 @@ gdm_display_add_user_authorization (GdmDisplay *self,
         /* Remove access for the programs run by greeter now that the
          * user session is starting.
          */
-        setup_xhost_auth (host_entries, si_entries);
-        gdm_error_trap_push ();
+        setup_xhost_auth (host_entries);
+
         for (i = 0; i < G_N_ELEMENTS (host_entries); i++) {
-                XRemoveHost (self->priv->x11_display, &host_entries[i]);
+                cookies[i] = xcb_change_hosts_checked (self->priv->xcb_connection,
+                                                       XCB_HOST_MODE_DELETE,
+                                                       host_entries[i].family,
+                                                       host_entries[i].length,
+                                                       (uint8_t *) host_entries[i].address);
         }
-        XSync (self->priv->x11_display, False);
-        if (gdm_error_trap_pop ()) {
-                g_warning ("Failed to remove greeter program access to the display. Trying to proceed.");
+
+        for (i = 0; i < G_N_ELEMENTS (cookies); i++) {
+                xcb_generic_error_t *xcb_error;
+
+                xcb_error = xcb_request_check (self->priv->xcb_connection, cookies[i]);
+
+                if (xcb_error != NULL) {
+                        g_warning ("Failed to remove greeter program access to the display. Trying to proceed.");
+                        free (xcb_error);
+                }
         }
 
         return TRUE;
@@ -649,6 +645,41 @@ gdm_display_finish (GdmDisplay *self)
         return TRUE;
 }
 
+static void
+gdm_display_disconnect (GdmDisplay *self)
+{
+        /* These 3 bits are reserved/unused by the X protocol */
+        guint32 unused_bits = 0b11100000000000000000000000000000;
+        XID highest_client, client;
+        guint32 client_increment;
+        const xcb_setup_t *setup;
+
+        if (self->priv->xcb_connection == NULL) {
+                return;
+        }
+
+        setup = xcb_get_setup (self->priv->xcb_connection);
+
+        /* resource_id_mask is the bits given to each client for
+         * addressing resources */
+        highest_client = (XID) ~unused_bits & ~setup->resource_id_mask;
+        client_increment = setup->resource_id_mask + 1;
+
+        /* Kill every client but ourselves, then close our own connection
+         */
+        for (client = 0;
+             client <= highest_client;
+             client += client_increment) {
+
+                if (client != setup->resource_id_base)
+                        xcb_kill_client (self->priv->xcb_connection, client);
+        }
+
+        xcb_flush (self->priv->xcb_connection);
+
+        g_clear_pointer (&self->priv->xcb_connection, xcb_disconnect);
+}
+
 gboolean
 gdm_display_unmanage (GdmDisplay *self)
 {
@@ -658,6 +689,8 @@ gdm_display_unmanage (GdmDisplay *self)
 
         g_debug ("GdmDisplay: unmanage display");
 
+        gdm_display_disconnect (self);
+
         g_timer_stop (self->priv->server_timer);
 
         if (self->priv->user_access_file != NULL) {
@@ -943,7 +976,7 @@ gdm_display_get_property (GObject        *object,
                 g_value_set_boolean (value, self->priv->is_local);
                 break;
         case PROP_IS_CONNECTED:
-                g_value_set_boolean (value, self->priv->x11_display != NULL);
+                g_value_set_boolean (value, self->priv->xcb_connection != NULL);
                 break;
         case PROP_LAUNCH_ENVIRONMENT:
                 g_value_set_object (value, self->priv->launch_environment);
@@ -1070,7 +1103,7 @@ register_display (GdmDisplay *self)
         if (self->priv->connection == NULL) {
                 g_critical ("error getting system bus: %s", error->message);
                 g_error_free (error);
-                exit (1);
+                exit (EXIT_FAILURE);
         }
 
         self->priv->object_skeleton = g_dbus_object_skeleton_new (self->priv->id);
@@ -1187,7 +1220,7 @@ gdm_display_class_init (GdmDisplayClass *klass)
                                                               "x11-display-name",
                                                               "x11-display-name",
                                                               NULL,
-                                                              G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+                                                              G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
         g_object_class_install_property (object_class,
                                          PROP_SEAT_ID,
                                          g_param_spec_string ("seat-id",
@@ -1422,10 +1455,129 @@ can_create_environment (const char *session_id)
         return session_exists;
 }
 
+#define ALREADY_RAN_INITIAL_SETUP_ON_THIS_BOOT GDM_RUN_DIR "/gdm.ran-initial-setup"
+
+static gboolean
+already_done_initial_setup_on_this_boot (void)
+{
+        GError *error = NULL;
+
+        if (g_file_test (ALREADY_RAN_INITIAL_SETUP_ON_THIS_BOOT, G_FILE_TEST_EXISTS))
+                return TRUE;
+
+        if (!g_file_set_contents (ALREADY_RAN_INITIAL_SETUP_ON_THIS_BOOT,
+                                  "1",
+                                  1,
+                                  &error)) {
+                g_warning ("GdmDisplay: Could not write initial-setup-done marker to %s: %s",
+                           ALREADY_RAN_INITIAL_SETUP_ON_THIS_BOOT,
+                           error->message);
+                g_clear_error (&error);
+        }
+
+        return FALSE;
+}
+
+static gboolean
+kernel_cmdline_initial_setup_argument (const gchar  *contents,
+                                       gchar       **initial_setup_argument,
+                                       GError      **error)
+{
+        GRegex *regex = NULL;
+        GMatchInfo *match_info = NULL;
+        gchar *match_group = NULL;
+
+        g_return_val_if_fail (initial_setup_argument != NULL, FALSE);
+
+        regex = g_regex_new ("\\bgnome.initial-setup=([^\\s]*)\\b", 0, 0, error);
+
+        if (!regex)
+            return FALSE;
+
+        if (!g_regex_match (regex, contents, 0, &match_info)) {
+                g_free (match_info);
+                g_free (regex);
+
+                g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
+                             "Could not match gnome.initial-setup= in kernel cmdline");
+
+                return FALSE;
+        }
+
+        match_group = g_match_info_fetch (match_info, 1);
+
+        if (!match_group) {
+                g_free (match_info);
+                g_free (regex);
+
+                g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
+                             "Could not match gnome.initial-setup= in kernel cmdline");
+
+                return FALSE;
+        }
+
+        *initial_setup_argument = match_group;
+
+        g_free (match_info);
+        g_free (regex);
+
+        return TRUE;
+}
+
+/* Function returns true if we had a force state in the kernel
+ * cmdline */
+static gboolean
+kernel_cmdline_initial_setup_force_state (gboolean *force_state)
+{
+        GError *error = NULL;
+        gchar *contents = NULL;
+        gchar *setup_argument = NULL;
+
+        g_return_val_if_fail (force_state != NULL, FALSE);
+
+        if (!g_file_get_contents ("/proc/cmdline", &contents, NULL, &error)) {
+                g_debug ("GdmDisplay: Could not check kernel parameters, not forcing initial setup: %s",
+                          error->message);
+                g_clear_error (&error);
+                return FALSE;
+        }
+
+        g_debug ("GdmDisplay: Checking kernel command buffer %s", contents);
+
+        if (!kernel_cmdline_initial_setup_argument (contents, &setup_argument, &error)) {
+                g_debug ("GdmDisplay: Failed to read kernel commandline: %s", error->message);
+                g_clear_pointer (&contents, g_free);
+                return FALSE;
+        }
+
+        g_clear_pointer (&contents, g_free);
+
+        /* Poor-man's check for truthy or falsey values */
+        *force_state = setup_argument[0] == '1';
+
+        g_free (setup_argument);
+        return TRUE;
+}
+
 static gboolean
 wants_initial_setup (GdmDisplay *self)
 {
         gboolean enabled = FALSE;
+        gboolean forced = FALSE;
+
+        if (already_done_initial_setup_on_this_boot ()) {
+                return FALSE;
+        }
+
+        if (kernel_cmdline_initial_setup_force_state (&forced)) {
+                if (forced) {
+                        g_debug ("GdmDisplay: Forcing gnome-initial-setup");
+                        return TRUE;
+                }
+
+                g_debug ("GdmDisplay: Forceing no gnome-initial-setup");
+                return FALSE;
+        }
 
         /* don't run initial-setup on remote displays
          */
@@ -1593,70 +1745,78 @@ gdm_display_stop_greeter_session (GdmDisplay *self)
         }
 }
 
+static xcb_window_t
+get_root_window (xcb_connection_t *connection,
+                 int               screen_number)
+{
+        xcb_screen_t *screen = NULL;
+        xcb_screen_iterator_t iter;
+
+        iter = xcb_setup_roots_iterator (xcb_get_setup (connection));
+        while (iter.rem) {
+                if (screen_number == 0)
+                        screen = iter.data;
+                screen_number--;
+                xcb_screen_next (&iter);
+        }
+
+        if (screen != NULL) {
+                return screen->root;
+        }
+
+        return XCB_WINDOW_NONE;
+}
+
 static void
 gdm_display_set_windowpath (GdmDisplay *self)
 {
         /* setting WINDOWPATH for clients */
-        Atom prop;
-        Atom actualtype;
-        int actualformat;
-        unsigned long nitems;
-        unsigned long bytes_after;
-        unsigned char *buf;
+        xcb_intern_atom_cookie_t atom_cookie;
+        xcb_intern_atom_reply_t *atom_reply = NULL;
+        xcb_get_property_cookie_t get_property_cookie;
+        xcb_get_property_reply_t *get_property_reply = NULL;
+        xcb_window_t root_window = XCB_WINDOW_NONE;
         const char *windowpath;
         char *newwindowpath;
-        unsigned long num;
+        uint32_t num;
         char nums[10];
         int numn;
 
-        prop = XInternAtom (self->priv->x11_display, "XFree86_VT", False);
-        if (prop == None) {
+        atom_cookie = xcb_intern_atom (self->priv->xcb_connection, 0, strlen("XFree86_VT"), "XFree86_VT");
+        atom_reply = xcb_intern_atom_reply (self->priv->xcb_connection, atom_cookie, NULL);
+
+        if (atom_reply == NULL) {
                 g_debug ("no XFree86_VT atom\n");
-                return;
+                goto out;
         }
-        if (XGetWindowProperty (self->priv->x11_display,
-                DefaultRootWindow (self->priv->x11_display), prop, 0, 1,
-                False, AnyPropertyType, &actualtype, &actualformat,
-                &nitems, &bytes_after, &buf)) {
+
+        root_window = get_root_window (self->priv->xcb_connection,
+                                       self->priv->xcb_screen_number);
+
+        if (root_window == XCB_WINDOW_NONE) {
+                g_debug ("couldn't find root window\n");
+                goto out;
+        }
+
+        get_property_cookie = xcb_get_property (self->priv->xcb_connection,
+                                                FALSE,
+                                                root_window,
+                                                atom_reply->atom,
+                                                XCB_ATOM_INTEGER,
+                                                0,
+                                                1);
+
+        get_property_reply = xcb_get_property_reply (self->priv->xcb_connection, get_property_cookie, NULL);
+
+        if (get_property_reply == NULL) {
                 g_debug ("no XFree86_VT property\n");
-                return;
+                goto out;
         }
 
-        if (nitems != 1) {
-                g_debug ("%lu items in XFree86_VT property!\n", nitems);
-                XFree (buf);
-                return;
-        }
-
-        switch (actualtype) {
-        case XA_CARDINAL:
-        case XA_INTEGER:
-        case XA_WINDOW:
-                switch (actualformat) {
-                case  8:
-                        num = (*(uint8_t  *)(void *)buf);
-                        break;
-                case 16:
-                        num = (*(uint16_t *)(void *)buf);
-                        break;
-                case 32:
-                        num = (*(long *)(void *)buf);
-                        break;
-                default:
-                        g_debug ("format %d in XFree86_VT property!\n", actualformat);
-                        XFree (buf);
-                        return;
-                }
-                break;
-        default:
-                g_debug ("type %lx in XFree86_VT property!\n", actualtype);
-                XFree (buf);
-                return;
-        }
-        XFree (buf);
+        num = ((uint32_t *) xcb_get_property_value (get_property_reply))[0];
 
         windowpath = getenv ("WINDOWPATH");
-        numn = snprintf (nums, sizeof (nums), "%lu", num);
+        numn = snprintf (nums, sizeof (nums), "%u", num);
         if (!windowpath) {
                 newwindowpath = malloc (numn + 1);
                 sprintf (newwindowpath, "%s", nums);
@@ -1666,11 +1826,15 @@ gdm_display_set_windowpath (GdmDisplay *self)
         }
 
         g_setenv ("WINDOWPATH", newwindowpath, TRUE);
+out:
+        g_clear_pointer (&atom_reply, free);
+        g_clear_pointer (&get_property_reply, free);
 }
 
 gboolean
 gdm_display_connect (GdmDisplay *self)
 {
+        xcb_auth_info_t *auth_info = NULL;
         gboolean ret;
 
         ret = FALSE;
@@ -1679,21 +1843,26 @@ gdm_display_connect (GdmDisplay *self)
 
         /* Get access to the display independent of current hostname */
         if (self->priv->x11_cookie != NULL) {
-                XSetAuthorization ("MIT-MAGIC-COOKIE-1",
-                                   strlen ("MIT-MAGIC-COOKIE-1"),
-                                   (gpointer)
-                                   self->priv->x11_cookie,
-                                   self->priv->x11_cookie_size);
+                auth_info = g_alloca (sizeof (xcb_auth_info_t));
+
+                auth_info->namelen = strlen ("MIT-MAGIC-COOKIE-1");
+                auth_info->name = "MIT-MAGIC-COOKIE-1";
+                auth_info->datalen = self->priv->x11_cookie_size;
+                auth_info->data = self->priv->x11_cookie;
+
         }
 
-        self->priv->x11_display = XOpenDisplay (self->priv->x11_display_name);
+        self->priv->xcb_connection = xcb_connect_to_display_with_auth_info (self->priv->x11_display_name,
+                                                                            auth_info,
+                                                                            &self->priv->xcb_screen_number);
 
-        if (self->priv->x11_display == NULL) {
+        if (xcb_connection_has_error (self->priv->xcb_connection)) {
+                g_clear_pointer (&self->priv->xcb_connection, xcb_disconnect);
                 g_warning ("Unable to connect to display %s", self->priv->x11_display_name);
                 ret = FALSE;
         } else if (self->priv->is_local) {
-                XServerInterpretedAddress si_entries[3];
                 XHostAddress              host_entries[3];
+                xcb_void_cookie_t         cookies[3];
                 int                       i;
 
                 g_debug ("GdmDisplay: Connected to display %s", self->priv->x11_display_name);
@@ -1701,17 +1870,27 @@ gdm_display_connect (GdmDisplay *self)
 
                 /* Give programs access to the display independent of current hostname
                  */
-                setup_xhost_auth (host_entries, si_entries);
-
-                gdm_error_trap_push ();
+                setup_xhost_auth (host_entries);
 
                 for (i = 0; i < G_N_ELEMENTS (host_entries); i++) {
-                        XAddHost (self->priv->x11_display, &host_entries[i]);
+                        cookies[i] = xcb_change_hosts_checked (self->priv->xcb_connection,
+                                                               XCB_HOST_MODE_INSERT,
+                                                               host_entries[i].family,
+                                                               host_entries[i].length,
+                                                               (uint8_t *) host_entries[i].address);
                 }
 
-                XSync (self->priv->x11_display, False);
-                if (gdm_error_trap_pop ()) {
-                        g_debug ("Failed to give some system users access to the display. Trying to proceed.");
+                for (i = 0; i < G_N_ELEMENTS (cookies); i++) {
+                        xcb_generic_error_t *xcb_error;
+
+                        xcb_error = xcb_request_check (self->priv->xcb_connection, cookies[i]);
+
+                        if (xcb_error != NULL) {
+                                g_debug ("Failed to give system user '%s' access to the display. Trying to proceed.", host_entries[i].address + sizeof ("localuser"));
+                                free (xcb_error);
+                        } else {
+                                g_debug ("Gave system user '%s' access to the display.", host_entries[i].address + sizeof ("localuser"));
+                        }
                 }
 
                 gdm_display_set_windowpath (self);
